        .file "rag_add_n.s"
//
//  INTEL CONFIDENTIAL
//
//  Intel intends to release this code under the LGPL license
//  as described below, but this code is still under development and testing.
//
//  Please report any problems with this function to the author at the email
//  address below or (503) 264-6622.
//
//  The easiest way to use this file is to save it as rag_add_n.s,
//  then assemble via ias rag_add_n.s
//  and finally link with the rag_add_n.o object file
//  before the -lgmp switch on the link command line.
//  Note: you may have to run the saved source though dos2unix.
//
//
//  Copyright 2004 Intel, Corp
//  This version (14-SEPT-2004) was translated by Roger A Golliver,
//  roger.a.golliver@intel.com, for the Itanium(tm)-2 microarchitecture.
//
//  IA-64 mpn_add_n -- Add two limb vectors of the same length > 0 and store
//  sum in a third limb vector.
//
//  Copyright 2000, 2001, 2002 Free Software Foundation, Inc.
//
//  This file is part of the GNU MP Library.
//
//  The GNU MP Library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//
//  The GNU MP Library is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
//  License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with the GNU MP Library; see the file COPYING.LIB.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//  MA 02111-1307, USA.
//
/////////////////// beginning of algorithm expressed in C /////////////////
//
//  IA-64 mpn_add_n -- Add two limb vectors of the same length > 0 and store
//  sum in a third limb vector.  Return the most significant limb of the
//  product, adjusted for carry-out from the addition.
//
// #include <ia64intrin.h>
//
// #include "gmp.h"
// #include "gmp-impl.h"
// #include "longlong.h"
//
// #if GMP_NAIL_BITS == 0
//
// mp_limb_t
// mpn_add_n (mp_ptr outWP, mp_srcptr inUP, mp_srcptr_inVP, mp_size_t inN)
// {
//   mp_limb_t lul, lvl;               /* rotating greg's with integers in them */
//
//   mp_limb_t lwl;                         /* static greg's with integers in them */
//   mp_ptr wp = outWP;                     /* static wp pointer for load */
//   mp_ptr up = inUP;                      /* static up pointer for load */
//   mp_ptr vp = inVP;                      /* static vp pointer for store */
//
//   int CinYes = 0;                        /* rotating predicate register */
//   int CinNo  = !CinYes;                  /* rotating predicate register */
//
//   for(int i = 0; i < inN; i++) {
//      lul = *up++;                        /* ld8 lul = [up],8 */
//      lvl = *vp++;                        /* ld8 lul = [vp],8 */
//
//      if(CinYes)lwl = lul + lvl + 1;      /* add lpl=lul,lvl   */
//      if(CinNo )lwl = lul + lvl;          /* add lpl=lul,lvl,1 */
//      CinYes = CinYes?(lwl <= lul)        /* (CinYes) cmp.leu CinYes,CinNo=lwl,lpl */
//                     :(lwl <  lul);       /* (CinNo)  cmp.ltu CinYes,CinNo=lwl,lpl */
//      CinNo = !CinYes;                    /* magic of cmp instruction */
//
//      *wp++ = lwl;                        /* st8 [wp]=lwl,8 */
//   }                                      /* br.ctop */
//   if(CinYes) lwl = 1;
//   if(CinNo ) lwl = 0;
//
//   return lwl;
// }
//
//#else
//you lose
//#endif
////////////////////// end of algorithm expressed in C ////////////////

// INPUT PARAMETERS

outWP = r32 // parameter 1
inUP  = r33 // parameter 2
inVP  = r34 // parameter 4
inN   = r35 // parameter 3

// STATIC COPIES

savearlc  = r21
savepr    = r22
savearpfs = r23

lwl       = r8 // ABI return register

up        = r16
vp        = r17
wp        = r18
iN        = r19

// ROTATING COPIES

S0      = p16;    S1  = p17;  // Two stage software pipelined loop //

CoutNo  = p18; CinNo  = p19;  // Tricky, carry out rotates into carry in //
CoutYes = p20; CinYes = p21;  // Tricky, carry out rotates into carry in //


lul0 = r36; lul1 = r37;
lvl0 = r38; lvl1 = r39;

        .section .text
        .proc  __gmpn_add_n#
        .align 64
        .global __gmpn_add_n#

__gmpn_add_n:
 {   .mmi // 1 //          //       IN, LOCAL, OUT, ROTATING
        alloc    savearpfs = ar.pfs, 4, 4   ,   0, 8 // M2 // ABI stuff
        add      iN = -1,inN           // pre-decrement for ar.lc and br.ctop instruction
        mov      savepr = pr           // I0 // ABI stuff
 } { .mmi
        mov      wp = outWP            // copy to static register for stores with auto update
        mov      up = inUP             // copy to static register for loads with auto update
        mov      vp = inVP ;;          // copy to static register for loads with auto update
 } // outWP, inUP, inVP and inN are now all free
 {   .mmi // 2 //
        cmp.lt   p0,S0 = (2-1), iN     // special case inN <= 2 (iN <= 1)
        nop.m    0
        mov      savearlc = ar.lc      // I0 // ABI stuff
 } { .mmb
        mov      lul1  = 0
        mov      lvl1  = 0
  (S0)  br.dpnt.few .rag_scalar ;;     // B0 // scheduled 0 latency branch
 }
 {   .mmi // 3 //
        nop.m    0
        nop.m    0
        mov      ar.lc = iN ;;         // I0 // set up loop count register for br.ctop instruction
 }
 {   .mmi // 4 //
        nop.m    0
        nop.m    0
        mov      ar.ec = 2 ;;          // I0 // set up loop count register for br.ctop instruction
 }
 {   .mmi // 5 //
        nop.m    0
        nop.m    0
        mov pr.rot = 0x90000 ;;        // I0 // Tricky // S0 true and CinNo are true, the rest are false
 }
/////////////////// drop into software pipelined loop ///////////////////////
        .align 32
        .pred.rel "mutex", CinYes , CinNo
.rag_loop:
 {   .mmi // 6 //
  (S0)  ld8       lul0=[up],8                 // S0 //   // M0 // assume L1 hit for schedule
  (S0)  ld8       lvl0=[vp],8                 // S0 //   // M1 // assume L1 hit for schedule
        nop.i     0
 } { .mmi
(CinYes) add      lwl=lul1,lvl1,1             // S1 //   //    // scheduled 2 cycle latency for ld8
(CinNo ) add      lwl=lul1,lvl1               // S1 //   //    // scheduled 2 cycle latency for ld8
        nop.i     0 ;;
 }
 {   .mmi // 7 //
(CinYes) cmp.leu  CoutYes,CoutNo=lwl,lul1     // S1 //   //    // scheduled 1 cycle for add w/carry
(CinNo ) cmp.ltu  CoutYes,CoutNo=lwl,lul1     // S1 //   //    // scheduled 1 cycle for add wo/carry
        nop.i     0 
 } { .mib
  (S1)  st8       [wp]=lwl,8                  // S1 //   // M2 // scheduled 1 cycle for add (w/wo) carry
        nop.i     0
        br.ctop.sptk.few .rag_loop ;;                    // B0 // scheduled 0 latency branch
 }
/////////////////// drop out of software pipelined loop ///////////////////////
        .pred.rel "mutex", CinYes , CinNo
 {   .mmi //  5 + 2*2 + 2 * inN  //
 (CinYes) add     lwl=r0,r0,1                            // M0
 (CinNo ) add     lwl=r0,r0                              // M1
        mov       ar.lc=savearlc ;;                      // I0 // ABI stuff
 }
 {   .mib // 10 + 2 * inN //
        nop.m     0                                      // M0
        mov       pr=savepr,0x1003e                      // I0 // ABI stuff
        br.ret.sptk.many  b0 ;;                          // B0
 }
/////////////////// branch into small scalar code ///////////////////////
        .align 32
.rag_scalar:
 {   .mmi
        ld8       lul0=[up],8                            // M0 // 
        ld8       lvl0=[vp],8                            // M1 //
        cmp.eq    S1,p0 = (2-1),iN
 } { .mmi
        cmp.ne    CinYes,CinNo = r0,r0
        nop.m     0
        nop.i     0 ;;
 }
 {   .mmi
  (S1)  ld8       lul1=[up],8                            // M0 // 
  (S1)  ld8       lvl1=[vp],8                            // M1 //
        nop.i     0                                      // I0 //
 }
        .pred.rel "mutex", CinYes , CinNo
 {   .mmi
 (CinYes) add    lwl = lul0,lvl0,1                       // M2 //
 (CinNo ) add    lwl = lul0,lvl0                         // M3 //
        cmp.eq    S0,p0 = (1-1), iN ;;                   // I1 // S0 <== (1 == N)
 }
        .pred.rel "mutex", CinYes , CinNo
 {   .mmi
 (CinYes) cmp.leu CoutYes,CoutNo = lwl,lul0              // M0 //
 (CinNo ) cmp.ltu CoutYes,CoutNo = lwl,lul0              // M1 // 
        nop.i     0                                      // I0 //
 } {   .mmb
        st8       [wp]=lwl,8                             // M2 //
        nop.m     0                                      // M3 //
  (S0)  br.dptk.few  .rag_scalar_ret ;;                  // B0 //
 }
        .pred.rel "mutex", CoutYes, CoutNo
 {   .mmi
 (CoutYes) cmp.eq  CinYes,CinNo = r0,r0                  // M0 // // fake rotate
 (CoutYes) add    lwl = lul1,lvl1,1                      // M1 //
 (CoutNo ) add    lwl = lul1,lvl1 ;;                     // I0 //
 }
        .pred.rel "mutex", CinYes , CinNo
 {   .mmi
 (CinYes) cmp.leu CoutYes,CoutNo = lwl,lul1              // M0 //
 (CinNo ) cmp.ltu CoutYes,CoutNo = lwl,lul1              // M1 // 
        cmp.eq    S0,p0 = (2-1), iN                      // I0 // S0 <== (2 == N)
 } {   .mmb
        st8       [wp]=lwl,8                             // M2 //
        nop.m     0
  (S0)  br.dptk.few  .rag_scalar_ret ;;                  // B0 //
 }

        .align 32
.rag_scalar_ret:
        .pred.rel "mutex", CoutYes, CoutNo
 {   .mmb
 (CoutYes) add    lwl = r0,r0,1                          // M0 //
 (CoutNo ) add    lwl = r0,r0                            // M1 //
        nop.b     0
 } {  .mib
        nop.m     0
        mov       pr=savepr,0x1003e                      // I0 // ABI stuff
        br.ret.sptk.many  b0 ;;                          // B0
 }
        .endp
