        .file "rag_addmul_1.s"
//
//  INTEL CONFIDENTIAL
//
//  Intel intends to release this code under the LGPL license
//  as described below, but this code is still under development and testing.
//
//  Please report any problems with this function to the author at the email
//  address below or (503) 264-6622.
//
//  The easiest way to use this file is to save it as rag_addmul_1.s,
//  then assemble via ias rag_addmul_1.s
//  and finally link with the rag_addmul_1.o object file
//  before the -lgmp switch on the link command line.
//  Note: you may have to run the saved source though dos2unix.
//
//
//  Copyright 2004 Intel, Corp
//  This version (14-SEPT-2004) was translated by Roger A Golliver,
//  roger.a.golliver@intel.com, for the Itanium(tm)-2 microarchitecture.
//
//  IA-64 mpn_addmul_1 -- Multiply a limb vector with a limb and add the
//  result to a second limb vector.
//
//  Copyright 2000, 2001, 2002 Free Software Foundation, Inc.
//
//  This file is part of the GNU MP Library.
//
//  The GNU MP Library is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//
//  The GNU MP Library is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
//  License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with the GNU MP Library; see the file COPYING.LIB.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//  MA 02111-1307, USA.
//
/////////////////// beginning of algorithm expressed in C /////////////////
//
//  mpn_addmul_1 -- multiply the N long limb vector pointed to by UP by VL,
//  add the N least significant limbs of the product to the limb vector
//  pointed to by RP.  Return the most significant limb of the product,
//  adjusted for carry-out from the addition.
//
// #include <ia64intrin.h>
//
// #include "gmp.h"
// #include "gmp-impl.h"
// #include "longlong.h"
//
// #if GMP_NAIL_BITS == 0
//
// mp_limb_t
// mpn_addmul_1 (mp_ptr in_outRP, mp_srcptr inUP, mp_size_t inN, mp_limb_t inVL)
// {
//   mp_limb_t retval, rpl = 0;             /* static   greg's with integers in them */
//   mp_limb_t fvl;                         /* static   freg's with integers in them */
//   mp_limb_t cl, hpl, lpl;                /* rotating greg's with integers in them */
//   mp_limb_t fvl, ful, frl, flpl, fhpl;   /* rotating freg's with integers in them */
//
//   mp_ptr rpr = in_outRP;                 /* static rp pointer for load */
//   mp_ptr rps = in_outRP;                 /* static rp pointer for store */
//   mp_ptr up  = inUP;                     /* static up pointer for load */
//
//   int CinYes = 0;                        /* rotating predicate register */
//   int CinNo  = !CinYes;                  /* rotating predicate register */
//
//   ASSERT (CinYes != CinNo);                     /* mutially exclusive predicates */
//   ASSERT (sizeof(__m64) == sizeof(mp_limb_t));  /* both uint64_t */
//   ASSERT (sizeof(__m64) == sizeof(mp_size_t));  /* both uint64_t */
//   ASSERT (n >= 1);                              /* at least one element in the bignum */
//   ASSERT (MPN_SAME_OR_SEPARATE_P (rp, up, n));  /* identical arrays or don't overlap */
//
//   fvl = vl;                              /* setf.sig fvl=vl */
//
//   for(int i = 0; i < inN; i++) {
//      ful = *up++;                        /* ldf8 ful = [up ],8 */
//      frl = *rpr++;                       /* ldf8 frl = [rpr],8 */
//
//      flpl = _m64_xmalu(ful, fvl, frl);   /* xma.lu flpl = ful,fvl,frl */
//      fhpl = _m64_xmahu(ful, fvl, frl);   /* xma.hu fhpl = ful,fvl,frl */
//
//      lpl = flpl;                         /* getf.sig lpl = flpl */
//      hpl = fhpl;                         /* getf.sig hpl = fhpl */
//
//      ASSERT (CinYes != CinNo);           /* mutially exclusive predicates */
//      if(CinYes)rpl  = lpl + cl + 1;      /* (CinYes) add rpl = lpl, cl, 1 */
//      if(CinNo )rpl  = lpl + cl;          /* (CinNo ) add rpl = lpl, cl */
//      cl = hpl;                           /* magic of rotating register */
//      CinYes = CinYes?(rpl <= lpl)        /* (CinYes) cmp.leu CinYes,CinNo=rpl,lpl */
//                     :(rpl <  lpl);       /* (CinNo)  cmp.ltu CinYes,CinNo=rpl,lpl */
//      CinNo = !CinYes;                    /* magic of cmp instruction */
//      *rps++ = rpl;                       /* st8 [rps]=rpl,8 */
//   }                                      /* br.ctop */
//   ASSERT (CinYes != CinNo);              /* mutially exclusive predicates */
//   if(CinYes) retval = cl + 1;
//   if(CinNo ) retval = cl;
//
//   return retval;
// }
//
//#else
//you lose
//#endif
////////////////////// end of algorithm expressed in C ////////////////

// INPUT PARAMETERS

in_outRP = r32 // parameter 1
inUP     = r33 // parameter 2
inN      = r34 // parameter 3
inVL     = r35 // parameter 4

// STATIC COPIES

savearlc  = r21
savepr    = r22
savearpfs = r23

retval    = r8 // ABI return register

rpr       = r16
up        = r17
rps       = r18
rpl       = r19

fvl       = f6

// ROTATING COPIES

S0 = p16; S1 = p17; S2 = p18; S3 = p19; S4 = p20; S5 = p21; S6 = p22; S7 = p23;
S8 = p24; S9 = p25; Sa = p26; Sb = p27; 

CinN0 = p28; CinN1 = p29; CinN2 = p30; CinN3 = p31; CinN4 = p32; CinN5 = p33; CinN6 = p34; CinN7 = p35;
CinN8 = p36; CinN9 = p37; CinNa = p38; CinNb = p39;

CinY0 = p40; CinY1 = p41; CinY2 = p42; CinY3 = p43; CinY4 = p44; CinY5 = p45; CinY6 = p46; CinY7 = p47;
CinY8 = p48; CinY9 = p49; CinYa = p50; CinYb = p51;

fup0 = f32; fup1 = f33; fup2 = f34; fup3 = f35; fup4 = f36; fup5 = f37; fup6 = f38; fup7 = f39;
fup8 = f40; fup9 = f41; fupa = f42; fupb = f43;

frp0 = f44; frp1 = f45; frp2 = f46; frp3 = f47; frp4 = f48; frp5 = f49; frp6 = f50; frp7 = f51;
frp8 = f52; frp9 = f53; frpa = f54; frpb = f55;

flpl0 = f56; flpl1 = f57; flpl2 = f58; flpl3 = f59; flpl4 = f60; flpl5 = f61; flpl6 = f62; flpl7 = f63;
flpl8 = f64; flpl9 = f65; flpla = f66; flplb = f67;

fhpl0 = f68; fhpl1 = f69; fhpl2 = f70; fhpl3 = f71; fhpl4 = f72; fhpl5 = f73; fhpl6 = f74; fhpl7 = f75;
fhpl8 = f76; fhpl9 = f77; fhpla = f78; fhplb = f79;

lpl0 = r32; lpl1 = r33; lpl2 = r34; lpl3 = r35; lpl4 = r36; lpl5 = r37; lpl6 = r38; lpl7 = r39;
lpl8 = r40; lpl9 = r41; lpla = r42; lplb = r43;

hpl0 = r44; hpl1 = r45; hpl2 = r46; hpl3 = r47; hpl4 = r48; hpl5 = r49; hpl6 = r50; hpl7 = r51;
hpl8 = r52; hpl9 = r53; hpla = r54; hplb = r55;

cl = hpla              // Tricky, hpl rotates into cl

CoutYes = CinY8    // Tricky, 
CinYes  = CinY9    // Tricky, 

CoutNo  = CinN8    // Tricky, 
CinNo   = CinN9    // Tricky, 

        .section .text
        .proc  __gmpn_addmul_1#
        .align 64
        .global __gmpn_addmul_1#

__gmpn_addmul_1:
 {   .mmi // 1 //          //       IN, LOCAL, OUT, ROTATING
        alloc    savearpfs = ar.pfs, 4, 28   ,   0, 32 // M2 // ABI stuff
        setf.sig fvl = inVL            // fvl is loop invariant
        mov      savepr = pr           // I0 // ABI stuff
 } { .mmi
        mov      rps = in_outRP        // copy to static register for stores with auto update
        mov      rpr = in_outRP        // copy to static register for loads with auto update
        mov      up = inUP ;;          // copy to static register for loads with auto update
 }
 {   .mmi // 2 //
        cmp.lt	 p0,S0 = 2, inN        // special case inN <= 2
        nop.m	 0
        nop.i	 0
 } { .mmb
        nop.m    0
        nop.m    0
  (S0)  br.dptk.few .rag_scalar ;;     // B0 // scheduled 0 latency branch
 } // in_outRP, inVL and inUP are now free
{   .mmi // 3 //
        add      inN = -1, inN         // pre-decrement for ar.lc and br.ctop instruction
        nop.m    0
        mov      savearlc = ar.lc      // I0 // ABI stuff
 } { .mmi
        nop.m	 0
        nop.m    0
        nop.i    0 ;;
 }
 {   .mmi // 4 //
        nop.m	 0
        mov      rpl = r0              // initialize to zero, so will generate CinNo true while pipeline fills up
        mov      ar.lc=inN             // I0 // set up loop count register for br.ctop instruction
 } { .mmi
        mov      retval = r0
        mov      hpl1 = r0             // initialize to zero, this will rotate into cl on stage 9
        nop.i    0 ;;
 } // inN is now free, all input registers are scratch
 {   .mmi // 5 //
        mov      lpl9 = r0             // initialize to zero, so will generate CinNo true while pipeline fills up
        mov      lpl8 = r0             // initialize to zero, so will generate CinNo true while pipeline fills up
        mov      ar.ec = 10            // I0 // set up epilog count register for br.ctop instruction
 } { .mmi
        mov      lpl7 = r0             // initialize to zero, so will generate CinNo true while pipeline fills up
        mov      lpl6 = r0             // initialize to zero, so will generate CinNo true while pipeline fills up
        mov      lpl5 = r0 ;;          // initialize to zero, so will generate CinNo true while pipeline fills up
 }
 {   .mmi // 6 //
        mov      lpl4 = r0             // initialize to zero, so will generate CinNo true while pipeline fills up
        mov      lpl3 = r0             // initialize to zero, so will generate CinNo true while pipeline fills up
        mov pr.rot = 0x2000010000      // I0 // rotating predicate S0 and CinNo9 true, the rest are false
 } { .mmi
        mov      lpl2 = r0             // initialize to zero, so will generate CinNo true while pipeline fills up
        mov      lpl1 = r0             // initialize to zero, so will generate CinNo true while pipeline fills up
        mov      lpl0 = r0 ;;          // initialize to zero, so will generate CinNo true while pipeline fills up
 }
/////////////////// drop into software pipelined loop ///////////////////////
        .pred.rel "mutex", CinYes , CinNo
	.align 32
.rag_loop:
 {   .mfi // 7 //
  (S0)  ldf8      fup0=[up ],8                           // M0 // assume L2 hit schedule for 10 cycles
  (S4)  xma.lu    flpl4=fup4,fvl,frp4                    // F0 // scheduled 9/10 cycle latency for ldf8
 (CinYes) add     rpl=lpl9,cl,1                          // I0 // scheduled 6/7 cycle latency for getf.sig
 } { .mfi
  (S6)  getf.sig  lpl6=flpl6                             // M2 // scheduled 4 cycle latency for xma
  (S4)  xma.hu    fhpl4=fup4,fvl,frp4                    // F1 // scheduled 9/10 cycle latency for ldf8
 (CinNo ) add     rpl=lpl9,cl ;;                         // I1 // scheduled 6/7 cycle latency for getf.sig
 }
 {   .mmi // 8 //
  (S0)  ldf8      frp0=[rpr],8                           // M0 // assume L2 hit schedule for 9 cycles
  (S6)  getf.sig  hpl6=fhpl6                             // M2 // scheduled 5 cycle latency for xma
  (CinYes) cmp.leu CoutYes,CoutNo=rpl,lpl9               // I0 // scheduled 1 cycle for add w/carry
 } { .mib
  (S9)  st8       [rps]=rpl,8                            // M3 // scheduled 1 cycle for add (w/wo) carry
  (CinNo) cmp.ltu CoutYes,CoutNo=rpl,lpl9                // I1 // scheduled 1 cycle for add wo/carry
        br.ctop.sptk.few .rag_loop ;;                    // B0 // scheduled 0 latency branch
 }
/////////////////// drop out of software pipelined loop ///////////////////////
        .pred.rel "mutex", CinYes , CinNo
 {   .mmi //  7 + 2*10 + 2 * inN  //
 (CinYes) add retval=cl,r0,1                             // M0 // scheduled 7 cycle latency for getf.sig
 (CinNo ) add retval=cl,r0                               // M1
        mov       ar.lc=savearlc ;;                      // I0 // ABI stuff
 }
 {   .mib // 28 + 2 * inN //
        nop.m     0                                      // M0
        mov       pr=savepr,0x1003e                      // I0 // ABI stuff
        br.ret.sptk.many  b0 ;;                          // B0
 }
/////////////////// branch into small scalar code ///////////////////////
	.align 32
.rag_scalar:
 {   .mmi
        ldf8      fup0=[up ],8                          // M0 // 
        ldf8      frp0=[rpr],8                          // M1 //
        cmp.eq    S1,p0 = 2,inN ;;	                // S1 <==  (N == 2)
 }

 {   .mmi
  (S1)  ldf8      fup1=[up ],8                           // M0 // 
  (S1)  ldf8      frp1=[rpr],8                           // M1 //
        cmp.eq    S0,p0 = 1, inN ;;  		 	 // S0 <== (1 == N)
 }

 {   .mfi
        nop.m     0
        xma.lu    flpl0=fup0,fvl,frp0                    // F0 //
        nop.i     0
 } { .mfi
        nop.m     0
        xma.hu    fhpl0=fup0,fvl,frp0                    // F1 //
        nop.i     0 ;;
 }

 {   .mfi
        nop.m     0
  (S1)  xma.lu    flpl1=fup1,fvl,frp1                    // F0 //
        nop.i     0
 } { .mfi
        nop.m     0
  (S1)  xma.hu    fhpl1=fup1,fvl,frp1                    // F1 //
        nop.i     0 ;;
 }

 {   .mmi
        getf.sig  lpl0=flpl0                             // M2 //
        getf.sig  hpl0=fhpl0                             // M2 // stall
        nop.i     0 ;;
 }

 {   .mmi
  (S1)  getf.sig  lpl1=flpl1                             // M2 //
  (S1)  getf.sig  hpl1=fhpl1                             // M2 // stall
        nop.i     0 ;;
 }

 {   .mmb
        st8       [rps]=lpl0,8                           // M2 // no carry in for first
        mov       retval=hpl0                            // M0 // carry out for first
  (S0)  br.dptk.few  .rag_scalar_ret ;;                  // B0 //
 }

 {   .mmi
        add       lpl1=retval,lpl1
        nop.m     0
        cmp.eq    S0,p0 = 2, inN ;;		         // S0 <== (2 == N)
 }

 {   .mmi
        cmp.ltu   CoutYes,CoutNo=lpl1,retval
        mov       retval=hpl1                            // M0 //
        nop.i     0 ;;
 }

 {   .mmb
        st8       [rps]=lpl1,8                           // M2 // 
 (CoutYes) add    retval=1,retval                        // I0 //
  (S0)  br.dptk.few  .rag_scalar_ret ;;                  // B0 //
 }
	.align 32
.rag_scalar_ret:
 {   .mib
        nop.m     0
        mov       pr=savepr,0x1003e                      // I0 // ABI stuff
        br.ret.sptk.many  b0 ;;                          // B0
 }
        .endp
